-- Haskell 98 Standard Prelude

-- note: Num, Enum and Fractional are built into Chameleon

hconstraint (Eq, Ord, Bounded, Real, Integral,
	     Floating, RealFrac, RealFloat,
	     Functor, Monad,
	     Read, Show)

-- Eq class
extern (==) :: (Eq a) => a -> a -> Bool
extern (/=) :: (Eq a) => a -> a -> Bool

-- Ord class
rule Ord a ==> Eq a
extern compare :: (Ord a) => a -> a -> Ordering
extern (<)  :: (Ord a) => a -> a -> Bool 
extern (<=) :: (Ord a) => a -> a -> Bool 
extern (>=) :: (Ord a) => a -> a -> Bool
extern (>)  :: (Ord a) => a -> a -> Bool
extern max  :: (Ord a) => a -> a -> a
extern min  :: (Ord a) => a -> a -> a

-- Enum class
extern succ :: (Enum a) => a -> a
extern pred :: (Enum a) => a -> a
extern toEnum       :: (Enum a) => Int -> a
extern fromEnum     :: (Enum a) => a -> Int
extern enumFrom     :: (Enum a) => a -> [a]             -- [n..]
extern enumFromThen :: (Enum a) => a -> a -> [a]        -- [n,n'..]
extern enumFromTo   :: (Enum a) => a -> a -> [a]        -- [n..m]
extern enumFromThenTo   :: a -> a -> a -> [a]   -- [n,n'..m]

-- Bounded class
extern minBound :: (Bounded a) => a
extern maxBound :: (Bounded a) => a

-- Num class
rule Num a ==> Eq a, Show a
extern (+) :: (Num a) => a -> a -> a
extern (-) :: (Num a) => a -> a -> a
extern (*) :: (Num a) => a -> a -> a
extern negate :: (Num a) => a -> a
extern abs :: (Num a) => a -> a
extern signum :: (Num a) => a -> a
extern fromInteger :: (Num a) => Integer -> a

-- Real class
rule Real a ==> Num a, Ord a
extern toRational :: (Real a) => a -> Rational

extern Rational

-- Integral class
rule Integral a ==> Real a, Enum a
extern quot :: (Integral a) => a -> a -> a 
extern rem  :: (Integral a) => a -> a -> a   
extern div  :: (Integral a) => a -> a -> a
extern mod  :: (Integral a) => a -> a -> a
extern quotRem :: (Integral a) => a -> a -> (a,a) 
extern divMod  :: (Integral a) => a -> a -> (a,a)
extern toInteger :: (Integral a) => a -> Integer
		
-- Fractional class
rule Fractional a ==> Num a
extern (/) :: (Fractional a) => a -> a -> a
extern recip :: (Fractional a) => a -> a
extern fromRational :: (Fractional a) => Rational -> a


-- Floating class
rule Floating a ==> Fractional a
extern pi  :: (Floating a) => a
extern exp :: (Floating a) => a -> a
extern log :: (Floating a) => a -> a
extern sqrt :: (Floating a) => a -> a
extern (**) :: (Floating a) => a -> a -> a
extern logBase :: (Floating a) => a -> a -> a
extern sin :: (Floating a) => a -> a
extern cos :: (Floating a) => a -> a
extern tan :: (Floating a) => a -> a
extern asin :: (Floating a) => a -> a
extern acos :: (Floating a) => a -> a
extern atan :: (Floating a) => a -> a
extern sinh :: (Floating a) => a -> a
extern cosh :: (Floating a) => a -> a
extern tanh :: (Floating a) => a -> a
extern asinh :: (Floating a) => a -> a
extern acosh :: (Floating a) => a -> a
extern atanh :: (Floating a) => a -> a

-- class RealFrac a
rule RealFrac a ==> Real a, Fractional a
extern properFraction :: (RealFrac a, Integral b) => a -> (b,a)
extern truncate :: (RealFrac a, Integral b) => a -> b 
extern round :: (RealFrac a, Integral b) => a -> b
extern ceiling :: (RealFrac a, Integral b) => a -> b
extern floor :: (RealFrac a, Integral b) => a -> b

-- class RealFloat
rule RealFloat a ==> RealFrac a, Floating a
extern floatRadix       :: (RealFloat a) => a -> Integer
extern floatDigits      :: (RealFloat a) => a -> Int
extern floatRange       :: (RealFloat a) => a -> (Int,Int)
extern decodeFloat      :: (RealFloat a) => a -> (Integer,Int)
extern encodeFloat      :: (RealFloat a) => Integer -> Int -> a
extern exponent         :: (RealFloat a) => a -> Int
extern significand      :: (RealFloat a) => a -> a
extern scaleFloat       :: (RealFloat a) => Int -> a -> a
extern isNaN		:: (RealFloat a) => a -> Bool
extern isInfinite	:: (RealFloat a) => a -> Bool
extern isDenormalized	:: (RealFloat a) => a -> Bool
extern isNegativeZero	:: (RealFloat a) => a -> Bool
extern isIEEE		:: (RealFloat a) => a -> Bool
extern atan2            :: (RealFloat a) => a -> a -> a

-- numeric functions
extern subtract   :: (Num a) => a -> a -> a
extern even	   :: (Integral a) => a -> Bool
extern odd        :: (Integral a) => a -> Bool

extern gcd              :: (Integral a) => a -> a -> a
extern lcm              :: (Integral a) => a -> a -> a
extern (^)              :: (Num a, Integral b) => a -> b -> a
extern (^^)             :: (Fractional a, Integral b) => a -> b -> a
extern fromIntegral     :: (Integral a, Num b) => a -> b
extern realToFrac     :: (Real a, Fractional b) => a -> b


-- Monadic classes

-- Functor class
extern fmap :: (Functor f) => (a -> b) -> f a -> f b

-- Monad class
extern (>>=)  :: (Monad m) => m a -> (a -> m b) -> m b
extern (>>)   :: (Monad m) => m a -> m b -> m b
extern return :: (Monad m) => a -> m a
extern fail   :: (Monad m) => [Char] -> m a

extern sequence       :: (Monad m) => [m a] -> m [a] 
extern sequence_      :: (Monad m) => [m a] -> m () 
extern mapM             :: (Monad m) => (a -> m b) -> [a] -> m [b]
extern mapM_            :: (Monad m) => (a -> m b) -> [a] -> m ()
extern (=<<)            :: (Monad m) => (a -> m b) -> m a -> m b


extern id    :: a -> a
extern const :: a -> b -> a
extern (.)   :: (b -> c) -> (a -> b) -> a -> c
extern flip  :: (a -> b -> c) -> b -> a -> c
extern seq  :: a -> b -> b
extern ($)  :: (a -> b) -> a -> b
extern ($!) :: (a -> b) -> a -> b

extern (&&) :: Bool -> Bool -> Bool
extern (||) :: Bool -> Bool -> Bool
extern not :: Bool -> Bool
-- extern otherwise :: Bool

-- Bool instances
rule Eq Bool <==> True
rule Ord Bool <==> True
rule Enum Bool <==> True
rule Bounded Bool <==> True


-- Char instances
rule Eq Char <==> True
rule Ord Char <==> True
rule Enum Char <==> True
rule Bounded Char <==> True

-- Maybe
extern Maybe
extern Nothing :: Maybe a
extern Just :: a -> Maybe a
rule Eq (Maybe a) <==> Eq a
rule Ord (Maybe a) <==> Ord a
rule Show (Maybe a) <==> Show a
rule Read (Maybe a) <==> Read a
rule Functor Maybe <==> True
rule Monad Maybe <==> True
extern maybe :: b -> (a -> b) -> Maybe a -> b

-- Either
extern Either
extern Left  :: a -> Either a b
extern Right :: b -> Either a b
rule Eq (Either a b) <==> Eq a, Eq b
rule Ord  (Either a b) <==> Ord a, Ord b
rule Show (Either a b) <==> Show a, Show b
rule Read (Either a b) <==> Read a, Read b
extern either :: (a -> c) -> (b -> c) -> Either a b -> c

-- IO
extern IO
rule Functor IO <==> True
rule Monad IO   <==> True

-- Ordering
extern Ordering
extern LT :: Ordering
extern EQ :: Ordering
extern GT :: Ordering

-- Numeric types
rule  Eq       Int <==> True
rule  Ord      Int <==> True
rule  Num      Int <==> True
rule  Real     Int <==> True
rule  Integral Int <==> True
rule  Enum     Int <==> True
rule  Bounded  Int <==> True

extern Integer
rule  Eq       Integer <==> True
rule  Ord      Integer <==> True
rule  Num      Integer <==> True
rule  Real     Integer <==> True
rule  Integral Integer <==> True
rule  Enum     Integer <==> True

extern Float
rule  Eq         Float <==> True
rule  Ord        Float <==> True
rule  Num        Float <==> True
rule  Real       Float <==> True
rule  Fractional Float <==> True
rule  Floating   Float <==> True
rule  RealFrac   Float <==> True
rule  RealFloat  Float <==> True

extern Double
rule  Eq         Double <==> True
rule  Ord        Double <==> True
rule  Num        Double <==> True
rule  Real       Double <==> True
rule  Fractional Double <==> True
rule  Floating   Double <==> True
rule  RealFrac   Double <==> True
rule  RealFloat  Double <==> True

rule  Enum Float  <==> True
rule  Enum Double <==> True

extern numericEnumFrom         :: (Fractional a) => a -> [a]
extern numericEnumFromThen     :: (Fractional a) => a -> a -> [a]
extern numericEnumFromTo       :: (Fractional a, Ord a) => a -> a -> [a]
extern numericEnumFromThenTo   :: (Fractional a, Ord a) => a -> a -> a -> [a]


-- Lists
rule Functor [] <==> True 	-- FIXME: parse error
rule Monad []   <==> True
rule Eq [a]  <==> Eq a
rule Ord [a] <==> Ord a

-- Tuples
extern fst              :: (a,b) -> a
extern snd              :: (a,b) -> b
extern curry            :: ((a, b) -> c) -> a -> b -> c
extern uncurry          :: (a -> b -> c) -> ((a, b) -> c)

-- Misc
extern until            :: (a -> Bool) -> (a -> a) -> a -> a
extern asTypeOf         :: a -> a -> a
extern error            :: [Char] -> a
--extern undefined        :: a


--------------------------------------------------------------------------------
-- PreludeList

extern map :: (a -> b) -> [a] -> [b]
extern (++) :: [a] -> [a] -> [a]
extern filter :: (a -> Bool) -> [a] -> [a]
extern concat :: [[a]] -> [a]
extern concatMap :: (a -> [b]) -> [a] -> [b]
extern head             :: [a] -> a
extern tail             :: [a] -> [a]
extern last             :: [a] -> a
extern init             :: [a] -> [a]
extern null             :: [a] -> Bool
extern length           :: [a] -> Int
extern (!!)                :: [a] -> Int -> a
extern foldl            :: (a -> b -> a) -> a -> [b] -> a
extern foldl1           :: (a -> a -> a) -> [a] -> a
extern scanl            :: (a -> b -> a) -> a -> [b] -> [a]
extern scanl1           :: (a -> a -> a) -> [a] -> [a]
extern foldr            :: (a -> b -> b) -> b -> [a] -> b
extern foldr1           :: (a -> a -> a) -> [a] -> a
extern scanr             :: (a -> b -> b) -> b -> [a] -> [b]
extern scanr1          :: (a -> a -> a) -> [a] -> [a]
extern iterate          :: (a -> a) -> a -> [a]
extern repeat           :: a -> [a]
extern replicate        :: Int -> a -> [a]
extern cycle            :: [a] -> [a]
extern take                   :: Int -> [a] -> [a]
extern drop                   :: Int -> [a] -> [a]
extern splitAt                  :: Int -> [a] -> ([a],[a])
extern takeWhile               :: (a -> Bool) -> [a] -> [a]
extern dropWhile               :: (a -> Bool) -> [a] -> [a]
extern span		  :: (a -> Bool) -> [a] -> ([a],[a])
extern break             :: (a -> Bool) -> [a] -> ([a],[a])
extern lines            :: [Char] -> [[Char]]
extern words            :: [Char] -> [[Char]]
extern unlines          :: [[Char]] -> [Char]
extern unwords          :: [[Char]] -> [Char]
extern reverse          :: [a] -> [a]
extern or          :: [Bool] -> Bool
extern and          :: [Bool] -> Bool
extern all         :: (a -> Bool) -> [a] -> Bool
extern any         :: (a -> Bool) -> [a] -> Bool
extern elem    :: (Eq a) => a -> [a] -> Bool
extern notElem    :: (Eq a) => a -> [a] -> Bool
extern lookup           :: (Eq a) => a -> [(a,b)] -> Maybe b
extern product     :: (Num a) => [a] -> a
extern sum     :: (Num a) => [a] -> a
extern minimum :: (Ord a) => [a] -> a
extern maximum :: (Ord a) => [a] -> a
extern zip              :: [a] -> [b] -> [(a,b)]
extern zip3             :: [a] -> [b] -> [c] -> [(a,b,c)]
extern zipWith          :: (a->b->c) -> [a]->[b]->[c]
extern zipWith3         :: (a->b->c->d) -> [a]->[b]->[c]->[d]
extern unzip            :: [(a,b)] -> ([a],[b])
extern unzip3           :: [(a,b,c)] -> ([a],[b],[c])

---------------------------------------------------------------------------------- PreludeText

-- Read class
extern readsPrec :: (Read a) => Int -> [Char] -> [(a,[Char])]
extern readList :: (Read a) => [Char] -> [([a],[Char])]

-- Show class
extern showsPrec        :: (Show a) => Int -> a -> [Char] -> [Char]
extern show             :: (Show a) => a -> [Char] 
extern showList         :: (Show a) => [a] -> [Char]

extern reads      :: (Read a) => [Char] -> [(a,[Char])] 
extern shows      :: (Show a) => a -> [Char] -> [Char]
extern read       :: (Read a) => [Char] -> a
extern showChar   :: Char -> [Char] -> [Char]
extern showString :: [Char] -> [Char] -> [Char]
extern showParen  :: Bool -> ([Char] -> [Char]) -> [Char] -> [Char]
extern readParen  :: Bool -> ([Char] -> [(a,[Char])]) -> [Char] -> [(a,[Char])]
extern lex        :: [Char] -> [([Char],[Char])]

-- instances
rule Show Int <==> True
rule Read Int  <==> True
rule Show Integer  <==> True
rule Read Integer  <==> True
rule Show Float  <==> True 
rule Read Float  <==> True
rule Show Double  <==> True
rule Read Double  <==> True
rule Show ()  <==> True
rule Read () <==> True
rule Show Bool <==> True
rule Read Bool <==> True

rule Show Char  <==> True
rule Read Char  <==> True
rule Show [a] <==> Show a
rule Read [a] <==> Read a
rule Show (a,b) <==> Show a, Show b
rule Read (a,b) <==> Read a, Read b


--------------------------------------------------------------------------------
-- PreludeIO

extern IOError
rule Show IOError <==> True
rule Eq IOError <==> True

extern ioError    :: IOError -> IO a
extern userError  :: [Char] -> IOError
extern catch      :: IO a -> (IOError -> IO a) -> IO a
extern putChar    :: Char -> IO ()
extern putStr     :: [Char] -> IO ()
extern putStrLn   :: [Char] -> IO ()
extern print      :: (Show a) => a -> IO ()
extern getChar    :: IO Char
extern getLine    :: IO [Char]
extern getContents :: IO [Char]
extern interact    ::  ([Char] -> [Char]) -> IO ()
extern readFile   :: [Char] -> IO [Char]
extern writeFile  :: [Char] -> [Char] -> IO ()
extern appendFile :: [Char] -> [Char] -> IO ()
extern readIO   :: (Read a) => [Char] -> IO a
extern readLn :: (Read a) => IO a
